#include "Diff.h"
#include "Grid.h"
#include <algorithm>
#include <map>

using std::map;

Diff::Diff() :
    m_pInfo(0),
    m_cells()
{
}

Diff::~Diff()
{
    if (m_pInfo) {
        delete m_pInfo;
    }
}

void applyDiff(Grid* pGrid, Diff* pDiff)
{
    Grid& grid = *pGrid;
    Diff& diff = *pDiff;

    std::swap(grid.m_pInfo, diff.m_pInfo);

    for (size_t i = 0; i < diff.m_cells.size(); i++) {
        CellDiff& cell = diff.m_cells[i];
        std::swap(grid.cell(cell.m_pos), cell.m_cellState);
    }
}

void concatDiff(const Grid& grid, const Diff& a, Diff* pB)
{
    //printf("BEGIN concatDiff\n");
    Diff& b = *pB;

    map<Pos, CellState> cells;

    for (size_t i = 0; i < a.m_cells.size(); i++) {
        const CellDiff& cell = a.m_cells[i];
        cells[cell.m_pos] = cell.m_cellState;
        //printf("a: (%u, %u)\n", cell.m_pos.m_x, cell.m_pos.m_y);
    }

    for (size_t i = 0; i < b.m_cells.size(); i++) {
        const CellDiff& cell = b.m_cells[i];
        cells[cell.m_pos] = cell.m_cellState;
        //printf("b: (%u, %u)\n", cell.m_pos.m_x, cell.m_pos.m_y);
    }

    b.m_cells.clear();

    map<Pos, CellState>::const_iterator iter = cells.begin();
    map<Pos, CellState>::const_iterator end = cells.end();
    for (; iter != end; ++iter) {
        if (iter->second != grid.cell(iter->first)) {
            b.m_cells.push_back(CellDiff(iter->first,
                                         iter->second));
            //printf("c: (%u, %u)\n", iter->first.m_x, iter->first.m_y);
        }
    }
    //printf("END concatDiff\n");
}
