#include "Zone.h"
#include "Grid.h"
#include <vector>

using std::vector;

Zone::Zone() :
    m_lift(false),
    m_lambda(false)
{
}

Zone::~Zone()
{
}

ZoneMap::ZoneMap() :
    m_width(0),
    m_height(0),
    m_data(0),
    m_zones(),
    m_zoneCount(0)
{
}

ZoneMap::~ZoneMap()
{
    if (m_data) {
        delete[] m_data;
    }
}

void ZoneMap::resize(uint32_t width, uint32_t height)
{
    m_width = width;
    m_height = height;
    if (m_data) {
        delete[] m_data;
    }

    m_data = new ZoneId[(width+2) * (height+2)];

    for (uint32_t y = 0; y <= m_height+1; y++) {
        for (uint32_t x = 0; x <= m_width+1; x++) {
            cell(x, y) = -1;
        }
    }
}

void makeZoneMap(ZoneMap* pZoneMap, const Grid& grid)
{
    ZoneMap& zoneMap = *pZoneMap;
    uint32_t width = grid.m_width;
    uint32_t height = grid.m_height;

    zoneMap.resize(width, height);

    for (uint32_t y = 1; y <= height; y++) {
        for (uint32_t x = 1; x <= width; x++) {
            CellState cs = grid.cell(x, y);
            if (csIsWall(cs) ||
                csIsLift(cs)) {
                zoneMap.cell(x, y) = -2;
            }
        }
    }

    for (uint32_t y = 1; y <= height; y++) {
        for (uint32_t j = 0; j <= 1; j++) {
            uint32_t start = j == 0 ? 1 : width;
            uint32_t end = j == 0 ? width+1 : 0;
            int32_t step = j == 0 ? 1 : -1;
            for (uint32_t x = start; x != end; x+=step) {
                CellState cs = grid.cell(x, y);
                if (csIsRock(cs) &&
                    zoneMap.cell(x, y-1) == -2) {

                    CellState csUnder = grid.cell(x, y-1);

                    if (zoneMap.cell(x-1, y) == -2 &&
                        zoneMap.cell(x+1, y) == -2) {
                        zoneMap.cell(x, y) = -2;
                    }

                    if (zoneMap.cell(x-1, y) == -2) {
                        if (csIsWall(csUnder) ||
                            csIsLift(csUnder)) {
                            zoneMap.cell(x, y) = -2;
                        } else if (csIsRock(csUnder)) {
                            if (zoneMap.cell(x+1, y-1) == -2) {
                                zoneMap.cell(x, y) = -2;
                            }
                        }
                    }

                    if (zoneMap.cell(x+1, y) == -2) {
                        if (csIsWall(csUnder) ||
                            csIsLift(csUnder)) {
                            zoneMap.cell(x, y) = -2;
                        } else if (csIsRock(csUnder)) {
                            if (zoneMap.cell(x-1, y-1) == -2) {
                                zoneMap.cell(x, y) = -2;
                            }
                        }
                    }
                }
            }
        }
    }

    while (true) {
        Pos start;
        if (zoneMap.m_zoneCount == 0) {
            start = grid.m_pInfo->m_robotPos;
        } else {
            for (uint32_t i = 1; i <= 9; i++) {
                uint32_t target = grid.m_trampolines[i-1].m_target;
                if (target != 0) {
                    Pos trampolinePos = grid.m_trampolines[i-1].m_pos;
                    if (zoneMap.cell(trampolinePos) >= 0) {
                        Pos targetPos = grid.m_targets[target-1].m_pos;
                        if (zoneMap.cell(targetPos) < 0) {
                            start = targetPos;
                            break;
                        }
                    }
                }
            }
        }

        if (start == Pos()) {
            break;
        }

        ZoneId zoneId = (ZoneId)zoneMap.m_zoneCount;
        zoneMap.m_zoneCount++;
        zoneMap.m_zones.resize(zoneMap.m_zoneCount);
        Zone& zone = zoneMap.m_zones[zoneId];

        //zoneMap.cell(start) = zoneId;

        vector<Pos> oldCells;
        vector<Pos> newCells;

        oldCells.push_back(start);

        while (!oldCells.empty()) {
            for (size_t i = 0; i < oldCells.size(); i++) {
                Pos pos = oldCells[i];

                if (zoneMap.cell(pos) == -1) {
                    zoneMap.cell(pos) = zoneId;
                    uint32_t x = pos.m_x;
                    uint32_t y = pos.m_y;
                    newCells.push_back(Pos(x-1, y));
                    newCells.push_back(Pos(x+1, y));
                    newCells.push_back(Pos(x, y-1));
                    newCells.push_back(Pos(x, y+1));

                    if (csIsLambda(grid.cell(pos))) {
                        zone.m_lambda = true;
                    }
                    if (csIsHORock(grid.cell(pos)) &&
                        zoneMap.cell(pos) != -2) {
                        zone.m_lambda = true;
                    }
                }

                if (csIsLift(grid.cell(pos))) {
                    zone.m_lift = true;
                }
            }

            oldCells.clear();
            oldCells.swap(newCells);
        }
    }
}
