#include "Grid.h"
#include "Sim.h"
#include "Diff.h"
#include "Zone.h"
#include <stdio.h>
#include <termios.h>
#include <unistd.h>
#include <vector>

using std::vector;

int getch()
{
    struct termios oldt, newt;
    int ch;
    int stdin_fd = fileno(stdin);
    tcgetattr(stdin_fd, &oldt);
    newt = oldt;
    newt.c_lflag &= ~(ICANON | ECHO);
    tcsetattr(stdin_fd, TCSANOW, &newt);
    ch = getchar();
    tcsetattr(stdin_fd, TCSANOW, &oldt);
    return ch;
}

int main(int argc, char* argv[])
{
    if (argc != 2) {
        fprintf(stderr, "Usage: play <mapfile>\n");
        return 1;
    }

    FILE* f = fopen(argv[1], "r");
    if (!f) {
        fprintf(stderr, "Error opening map file\n");
        return 1;
    }

    Grid grid;
    if (!readGrid(f, &grid)) {
        fclose(f);
        fprintf(stderr, "Error in map\n");
        return 1;
    }

    fclose(f);

#if 0
    {
        ZoneMap zoneMap;
        makeZoneMap(&zoneMap, grid);

        for (uint32_t i = 0; i < zoneMap.m_zoneCount; i++) {
            const Zone& zone = zoneMap.m_zones[i];
            printf("%u:%s%s\n",
                   i,
                   zone.m_lift ? " Lift" : "",
                   zone.m_lambda ? " Lambda" : "");
        }

        for (uint32_t y = 1; y <= grid.m_height; y++) {
            for (uint32_t x = 1; x <= grid.m_width; x++) {
                if (zoneMap.cell(x, y) < 0) {
                    grid.cell(x, y) = CS_Wall;
                }
            }
        }
    }
#endif

    puts("\x1b[2J\x1b[H");
    writeGrid(stdout, grid, true);

    vector<Action> actions;
    vector<Diff*> diffs;

    vector<Action> redoActions;
    vector<Diff*> redoDiffs;

    bool esc = false;
    bool escLeftBracket = false;
    bool escO = false;
    while (grid.m_pInfo->m_state == State_Running || true) {
        int iCh = getch();
        char ch = (char)iCh;
        if (ch == 'q') {
            break;
        }

        bool gotAction = false;
        bool gotUndo = false;
        bool gotRedo = false;
        Action ac = Action_Wait;

        if (escLeftBracket) {
            escLeftBracket = false;
            switch (ch) {
            case 'A': gotAction = true; ac = Action_MoveUp; break;
            case 'B': gotAction = true; ac = Action_MoveDown; break;
            case 'C': gotAction = true; ac = Action_MoveRight; break;
            case 'D': gotAction = true; ac = Action_MoveLeft; break;
            }
        } else if (escO) {
            escO = false;
            switch (ch) {
            case 'C': gotRedo = true; break;
            case 'D': gotUndo = true; break;
            }
        } else if (esc) {
            esc = false;
            switch (ch) {
            case '[': escLeftBracket = true; break;
            case 'O': escO = true; break;
            }
        } else {
            switch (ch) {
            case '\x1b': esc = true; break;
            case 'U': case 'u': gotAction = true; ac = Action_MoveUp; break;
            case 'D': case 'd': gotAction = true; ac = Action_MoveDown; break;
            case 'L': case 'l': gotAction = true; ac = Action_MoveLeft; break;
            case 'R': case 'r': gotAction = true; ac = Action_MoveRight; break;
            case ' ':
            case 'W': case 'w': gotAction = true; ac = Action_Wait; break;
            case 'A': case 'a': gotAction = true; ac = Action_Abort; break;
            case 'S': case 's': gotAction = true; ac = Action_Shave; break;
            }
        }

        if (gotAction && grid.m_pInfo->m_state == State_Running) {
            redoActions.clear();
            redoDiffs.clear();

            actions.push_back(ac);
            Diff* pDiff = new Diff;
            diffs.push_back(pDiff);
            step(&grid, ac, pDiff);
            puts("\x1b[H");
            writeGrid(stdout, grid, true);
        }

        if (gotUndo) {
            if (!actions.empty()) {
                redoActions.push_back(actions.back());
                actions.pop_back();

                Diff *pDiff = diffs.back();
                diffs.pop_back();
                applyDiff(&grid, pDiff);
                redoDiffs.push_back(pDiff);

                puts("\x1b[H");
                writeGrid(stdout, grid, true);
            }
        }

        if (gotRedo) {
            if (!redoActions.empty()) {
                actions.push_back(redoActions.back());
                redoActions.pop_back();

                Diff *pDiff = redoDiffs.back();
                redoDiffs.pop_back();
                applyDiff(&grid, pDiff);
                diffs.push_back(pDiff);

                puts("\x1b[H");
                writeGrid(stdout, grid, true);
            }
        }

    }

    printf("\n");
    printf("Actions:\n");
    writeTrace(stdout, actions);
    printf("\n");

    for (size_t i = 0; i < diffs.size(); i++) {
        delete diffs[i];
    }

    return 0;
}
